/**********************************************************/
/**********************************************************/
/*  $Header$  */
/**********************************************************/
/*  This is where all the high level comms is done.       */
/*  This module sits on top of channel.c which does       */
/*  buffered asyncronous communcations between two        */
/*  processes.                                            */
/*                                                        */
/*  Ask another program to monitor objects, rather than   */
/*  doing it locally.                                     */
/*  Each monitor is specified by its address and by a     */
/*  pattern.  Rules added later have priority over those  */
/*  added earlier.                                        */
/**********************************************************/

#include "db/Entity.h"
#include "general/Channel.h"
#include <netdb.h>

/***************************************************************************/
/*  Lookup table, this should be some kind of lexical automaton, with !!!  */
/*  end nodes pointing to (host,port) or NOTFOUND.  As this is quite  !!!  */
/*  hard to do, a simple table will be used as a prototype.           !!!  */
/***************************************************************************/

struct redirect_rule {
	String pattern;
	String host;
	int port;
	Channel channel;
	Cardinal ping_id;
	RedirectRule next;
	unsigned long int last_checked;
	EntityList entitys;
};
#define MAXPINGID 65535
static RedirectRule RedirectList;
static Boolean Disable = False;
static EntityList BaseRule;

/****************************************************************************/
/*  List redirects */
/****************************************************************************/

extern void list_redirects ()
{
	RedirectRule rule;
	int n = 0;

	start_report ("Listing redirects");
	FORALL (rule, RedirectList, next) {
		++ n;
		write_report ("%d.", n);
		write_report ("\thost    [%s+%d]", rule->host, rule->port);
		if (rule->channel)
			writeln_report (" Connected");
		else
			writeln_report (" Not connected");
		writeln_report ("\tpattern [%s]", rule->pattern);
		writeln_report ("\tentities [%d]", sizeOf_entityList(rule->entitys));
	}
	end_report ();
}

/****************************************************************************/
/*  Handle messages sent from the pinger process  */
/*  !!! Not very efficient lookup, should have 'keyed' objectlists !!! */
/*  !!! Should update each system in one go, not for each attribute */
/*  !!! This should really be put in the 'status' library.                  */
/****************************************************************************/

static Boolean scan (message, id)
	String *message;
	Cardinal *id;
{
	String s = *message;

	while (!isdigit(*s)) {
		if (*s == NULL)
			return False;
		s++;
	}
	for (*id = 0; isdigit(*s); s++)
		*id = (*id) * 10 + (Cardinal) (*s - '0');
	*message = s;
	return True;
}

static Boolean handle_response (rule, result)
	RedirectRule rule;
	String result;
{
	Cardinal id;
	Entity o;

	if (rule == NULL) {
		sys_error ("handle_reponse: rule is NULL");
		return False; /* -- Disconnect */
	}

	/* Read message, convert it into a status message */
	/* Set the approriate status bits in the entity.     */
	/* Arrange for 'request' to return this status.   */
	/* Possibly relay this new information straight   */
	/* back the channels.                              */

	debug_message ("Received [%s] from redirect server", result);
	while (scan (&result, &id)) {
		if (*result == MonitorEndCode || *result == '\n')
			continue;
		do {
			FOR_ENTITYLIST (o, rule->entitys)
				if (PINGID(o) == id) {
					if (! Entity_decodeCompactAscii (o, &result))
						user_error ("Couldn't parse status [%s]", result);

				}
		} while (*result == MonitorSepCode);
		for (; *result != MonitorEndCode && *result != NULL; result++);
	}
	return True;
}

/***************************************************************************/
/*  When a channel disconnects we must set all the Redirects connected     */
/*  to it to NULL.  The redirects themselves are not destroyed, just       */
/*  disabled until a new channel can be created.                           */
/***************************************************************************/

static void disconnect (rule)
	RedirectRule rule;
{
	RedirectRule r;
	FORALL(r, RedirectList, next)
		if (r->channel == rule->channel)
			r->channel = NULL;
	user_inform ("Redirect Channel disconnected");
	Channel_free (rule->channel);
}

/***************************************************************************/

/***************************************************************************/

static void open_redirect_channel (rule)
	RedirectRule rule;
{
	RedirectRule l;

	if (rule->channel)
		return;
	FORALL(l, RedirectList, next)
		if (l != rule && EQS(rule->host, l->host) && rule->port==l->port) {
			rule->channel = l->channel;
			return;
		}
	rule->channel = Channel_create_client (
		rule->port, rule->host, handle_response, disconnect, (caddr_t) rule);
	if (rule->channel) {
		char buffer[512]; /* !!! */
		user_inform ("Channel created");
		(void) sprintf (buffer, "hello %s %u",
			Config_local_host, Config_pinger_port);
		Channel_send (rule->channel, buffer);
		Channel_end_of_record (rule->channel);
		Channel_flush (rule->channel);
	} else
		user_inform ("Channel not created");
}

/***************************************************************************/
/*  Initiate transfer  */
/***************************************************************************/

extern void Redirect_say_hello (rule)
	RedirectRule rule;
{
	if (rule && rule->channel) {
		char buffer[512]; /* !!! */
		(void) sprintf (buffer, "hello %s %u",
			Config_local_host, Config_pinger_port);
		Channel_send (rule->channel, buffer);
		Channel_end_of_record (rule->channel);
		Channel_flush (rule->channel);
	}
}

/***************************************************************************/
/*  Try to attach to any disconnected servers  */
/***************************************************************************/

extern void reconnect_redirects ()
{
	RedirectRule rule;

	if (Disable)
		return;
	FORALL (rule, RedirectList, next)
		open_redirect_channel (rule);
}		

/***************************************************************************/
/*  Temporarily close connections  */
/***************************************************************************/

extern void disable_redirects ()
{
	RedirectRule rule, l;

	FORALL (rule, RedirectList, next) {
		Channel channel = rule->channel;
		if (channel) {
			Channel_free (channel);
			FORALL (l, RedirectList, next)
				if (channel==l->channel)
					l->channel=NULL;
		}
	}
	Disable = True;
}

extern void enable_redirects ()
{
	if (Disable) {
		Disable = False;
		reconnect_redirects ();
	}
}

/********************************************************************/
/*  Check to see if a entity is redirected elsewhere.                  */
/*  !!! Should maybe use this routine to check for new matches.     */
/********************************************************************/

extern Boolean isRedirected (entity)
	Entity entity;
{
	if (entity->redirect == NULL)
		return False;
	if (entity->redirect->channel == NULL) {
		/*  A redirect applies to this object but there is no connection */
		return False;
	}
	return True;
}

/********************************************************************/
/*  Try to find a matching redirection rule                         */
/********************************************************************/

extern RedirectRule try_redirect (entity)
	Entity entity;
{
	char buffer[1024]; /* !!! */
	String title;
	RedirectRule rule;

/*	if (!(title = get_ping_instruction (entity)))
		return NULL;
*/
	title = Entity_title (entity);
	/* Match address against list of regular expressions */
	FORALL(rule, RedirectList, next) {
		debug_message ("Matching pattern [%s] against [%s]", rule->pattern, title);
		if (wildmat (title, rule->pattern))
			break;
	}
	if (entity->redirect == rule) {
		/* -- Already associated with this redirect */
		return rule;
	}
	if (rule == NULL) {
		/* -- Previous rule no longer applies, should this happen ? */
		stop_redirect (entity);
		(void) insert_entityList (BaseRule, entity);
		return NULL;
	}
	if (rule->ping_id == MAXPINGID) {
		/* !!! Should recover from this condition, ping id's ought to be long */
		user_error ("Sorry no more Ping ID's available, can't ping anymore entities");
		return False;
	}
	if (entity->redirect) {
		/* -- Previous rule no longer applies  */
		stop_redirect (entity);
	} else {
		(void) extract_entityList (BaseRule, entity);
	}
	PINGID(entity) = rule->ping_id;
	insert_entityList (rule->entitys, entity);
	entity->redirect = rule;

	(void) sprintf (buffer, "request%u %s", rule->ping_id, title);
	rule->ping_id ++;
	debug_message ("Sending Redirect Request [%s] to monitor %d",
		buffer, rule->channel);
	Channel_send (rule->channel, buffer);
	Channel_end_of_record (rule->channel);
	Channel_flush (rule->channel);

	return rule;
}

/******************************************************************/
/*  Stop redirecting requests for a entity  */
/******************************************************************/

extern Boolean stop_redirect (entity)
	Entity entity;
{
	(void) extract_entityList (BaseRule, entity);
	if (entity->redirect == NULL)
		return False;
	if (entity->redirect->channel != NULL) {
		char buffer [1024]; /* !!! */
		(void) sprintf (buffer, "delete%u", PINGID(entity));
		Channel_send (entity->redirect->channel, buffer);
		Channel_end_of_record (entity->redirect->channel);
	}
	(void) extract_entityList (entity->redirect->entitys, entity);
	PINGID(entity) = 0;
	entity->redirect = NULL;
	return True;
}

/****************************************************************************/
/*  Reset/Clear all redirects  */
/****************************************************************************/

extern void clear_redirects ()
{
	while (RedirectList != NULL)
		free_redirect (RedirectList);
}

/***************************************************************/

/***************************************************************/

extern void Redirect_get_info (rule,host, port, pattern)
	RedirectRule rule;
	String *host, *pattern;
	int *port;
{
	if ( ! rule) {
		sys_error ("rule is NULL");
		*host = NULL;
		*pattern = NULL;
		return;
	}
	*host    = rule->host;
	*pattern = rule->pattern;
	*port    = rule->port;
}

/*****************************************************************/
/*  Create basic redirect  */
/*****************************************************************/

static RedirectRule create_basic (host, port, pattern)
	String host, pattern;
	int port;
{
	RedirectRule new;

	new = MonetNew (struct redirect_rule);
	new -> pattern = MonetNewString (pattern);
	new -> host    = MonetNewString (host);
	new -> port    = port;
	new -> channel = NULL;
	new -> next    = RedirectList;
	new -> ping_id = 1;
	new -> entitys = create_entityList ();
	RedirectList   = new;

	/*  Should look through all entities to see if any match this redirect rule  */

	return new;
}

/***************************************************************/
/*  Add and remove a redirect patern  */
/***************************************************************/

extern RedirectRule create_redirect (host, port, pattern)
	String host, pattern;
	int port;
{
	RedirectRule new;
	struct hostent *hp;
	Entity entity;

	/* Try to connect to remote monitor and say hello */
	/* Add the regexp and new connection to a table   */

	if ((hp = gethostbyname (host)) == NULL) {
		user_error ("redirect: No such host [%s]", host);
		return NULL;
	}
	host = hp->h_name;

	if (new = create_basic (hp->h_name, port, pattern)) {
		/*  -- look through all entities in BaseRule to see if any match this redirect rule  */
		FOR_ENTITYLIST (entity, BaseRule)
			try_redirect (entity);
		/*  get a channel for this rule  */
		open_redirect_channel (new);
	}
	return new;
}

/*****************************************************************/
/*  Is a redirct actually connected to a channel  */
/*****************************************************************/

extern Boolean redirect_connected (rule)
	RedirectRule rule;
{
	return (rule->channel != NULL);
}

/*****************************************************************/
/*  Free a redirect rule                                         */
/*****************************************************************/

extern void free_redirect (rule)
	RedirectRule rule;
{
	Channel channel;
	RedirectRule r;
	Entity entity;

	debug_message ("Removing redirect [%s+%u]", rule->host, rule->port);
	channel = rule->channel;
	MonetFree (rule->pattern);
	MonetFree (rule->host);
	MonetFree (rule);

	/*  Notify server by clearing requests !!! */
	/*  !!! This is where an acknowledgement is needed, else */
	/*  !!! we could get answers to queries just deleted */
	if (rule->channel) {
		Channel_send (rule->channel, "clear");
		Channel_end_of_record (rule->channel);
	}

	/*  Remove 'rule' from 'RedirectList' */
	LIST_DEL (RedirectRule, rule, RedirectList, next);

	/*  Try to find an alternative redirect for each entity  */
	FOR_ENTITYLIST (entity, rule->entitys) {
		PINGID(entity) = 0;
		entity->redirect = NULL;
		try_redirect (entity);
	}
	reset_entityList (rule->entitys);
	free_entityList (rule->entitys);

	/* A convoluted way of only closing a connection to a server */
	/* when there are no redirects to it. */
	FORALL(r, RedirectList, next)
		if (r->channel == channel)
			return;
	Channel_free (channel);
}

/*****************************************************************/
/* Init and Finish */
/*****************************************************************/

extern void finish_redirect ()
{
	clear_redirects ();
	reset_entityList (BaseRule);
	free_entityList (BaseRule);
}

extern void init_redirect ()
{
	RedirectList = NULL;
	BaseRule = create_entityList ();
}

/**********************************************************/
/**********************************************************/
/*  $Header$  */
/**********************************************************/
/**********************************************************/

