/***************************************************************************/
/***************************************************************************/
/*  $Header$  */
/***************************************************************************/
/*  Provides an asyncrous buffered channel between two processes.          */
/*  An interface to sockets, could be improved to say the least    !!!     */
/*  Uses Non blocking sockets.                                             */
/***************************************************************************/

#include "General.h"
#include "Channel.h"
#include "NiceSocket.h"
#include <sys/types.h>
#include <errno.h>
#include <sys/ioctl.h>

/***************************************************************************/
/*  Private definitions  */
/*  Exports Server and Channel, these are used externally as identifiers to */
/*  connections  */
/***************************************************************************/

#define END_OF_RECORD  NULL
#define BUFFERLENGTH   (2<<13)

static Channel ChannelList = 0;

struct channel {
	caddr_t subclass_data;
	int fid;
	ChannelHandler handler;
	ChannelDisconnect disconnect;
	char buffer [BUFFERLENGTH];
	char * pbuffer;
	Channel next;
};

/***************************************************************************/
/*  Read one \n or NULL terminated line from a socket and catch errors     */
/* If it returns 'finish==True' an error occured or disconnected */
/* !!! Should check len */
/* The routine will return on finding an end of record marker or when the  */
/* socket would block (if is a non-blocking socket).  In the latter case   */
/* the user must provide buffering.  */
/* !!! This is a mess now, clean it up */
/***************************************************************************/

static String read_line (fid, text, len,  finish)
	int fid, len;
	String *text;
	Boolean *finish;
{
	int cc;
	char * r;
	Boolean max = False;

	PROC ("read_line");
	if (fid < 0) {
		sys_error ("fid [%d] is invalid", fid);
		*finish = True;
		return NULL;
	}
	*finish = False;
	for (r = (*text); True; r++) {
		if (r-(*text) >= len) {
			if (! max) {
				max = True;
				sys_error ("Max message length %d exceeded", len);
			}
			r --;  /* Hmmm !!! */
		}
		cc = recv (fid, r, 1, 0);
		if (cc < 0) {
			if (errno == EWOULDBLOCK) {
				*r = NULL;
				debug_message ("Socket %d would block, buffer [%s]", fid, *text);
				*text = r;
				*finish = False;
			 } else if (errno == ENOTCONN) {
				sys_error ("Socket is not connected");
				*finish = True;
			} else {
				sys_error ("Bad recv: errno [%d]", errno);
				*finish = True;
			}
			return NULL;
		} else if (cc == 0) {
			user_warning ("recv: cc == 0, channel seems to have disconnected");
			*finish = True;
			return NULL;
		} else 	if (cc != 1) {
			sys_error ("Odd cc= recv (%d,...), cc = [%d]", fid, cc);
			*finish = True;
			return NULL;
		} else if ((*r == '\0' || *r == '\n') && r != (*text)) {
			*finish = False;
			*r = NULL;   /* Null terminate the string */
			return (*text);
		} else if (!isprint (*r)) {
			/*  Non printing characters are converted into whitespace */
			/*  This is primarily to deal with the problem of \r when */
			/*  using telnet.  */
			*r = ' ';
		} else {
			/*  *r is a printable character */
		}
	};
	/* NOTREACHED */
}

/***************************************************************************/
/*  This is called whenever new data arrives from the server.  Reads in as */
/*  much data as it can, catches some errors and cleanly disconnects       */
/***************************************************************************/

static void HandleChannelMessage (fid, channel_data)
	int fid;
	caddr_t channel_data;
{
	int len;
	char * r;
	Channel channel;
	Boolean finish;

	PROC ("HandleChannelMessage");

	FORALL(channel, ChannelList, next)
		if (channel->fid == fid)
			break;
	if (channel == NULL) {
		sys_error ("Couldn't identify channel for fid %d", fid);
		return;
	}
	debug_message ("New Messages on connection %d", fid);

	len = sizeof(channel->buffer) - (channel->pbuffer - channel->buffer);
	while ((r = read_line (channel->fid, &(channel->pbuffer), len, &finish)) != NULL) {
		channel->pbuffer = channel->buffer;
		if (channel->handler)
			if ((channel->handler (channel_data, channel->buffer)) == False) {
				finish = True;
				break;
			}
	}
	debug_message ("End of messages on connection %d", fid);
	if (finish) {
		free_input_fid (channel->fid);
		if (channel -> disconnect != NULL)
			channel->disconnect (channel_data);
	}
}

/***************************************************************************/
/*  Create a new channel  */
/***************************************************************************/

extern Channel Channel_create (fid, result_handler, disconnect, data)
	int fid;
	ChannelHandler result_handler;
	ChannelDisconnect disconnect;
	caddr_t data;
{
	Channel channel;

	channel = MonetNew (struct channel);
	if (channel == NULL) {
		sys_error ("Could not allocate space for channel");
		return NULL;
	}
	channel->subclass_data = NULL;
	channel->next = ChannelList;
	ChannelList = channel;
	channel -> fid = fid;
	channel -> handler = result_handler;
	channel -> disconnect = disconnect;
	channel -> pbuffer = channel->buffer;
	(void) create_input_fid (fid, fid_read, HandleChannelMessage, (caddr_t) data);
	return channel;
}

/***************************************************************************/
/*  Used to start a new channel, supply address of server and callback      */
/*  routines for new messages and disconnection.                           */
/***************************************************************************/

extern Channel Channel_create_client (server_port, server_address, result_handler, disconnect, data)
	int server_port;
	String server_address;
	ChannelHandler result_handler;
	ChannelDisconnect disconnect;
	caddr_t data;
{
	Channel channel;
	int n = 1, fid;

	PROC ("init_channel");

	fid = clientsock (server_address, server_port);
	if (fid == -9999) {
		user_error ("Address of pinger [%s] is bad", server_address);
		return NULL_CHANNEL;
	} else if (fid < 0) {
		user_inform ("Can't connect to server at address [%s], is it running?", server_address);
		return NULL_CHANNEL;
	}	
	/* Non blocking */
	ioctl(fid, FIONBIO, &n);
	user_inform ("Connected to server [%s] port %d", server_address, server_port);

	channel = Channel_create (fid, result_handler, disconnect, data);

	/* Send an initial message to get things going, Hack !!! */

	return channel;
}

/***************************************************************************/
/*  Gets rid of a channels datastructure  */
/***************************************************************************/

extern void Channel_free (channel)
	Channel channel;
{
	if (channel) {
		LIST_DEL (Channel, channel, ChannelList, next);
		free_input_fid (channel->fid);
		MonetFree (channel);
	}
}

/***************************************************************************/
/*  Send a message from the channel to the server */
/*  Do not send null messages, this avoids problems with strlen too */
/***************************************************************************/

extern void Channel_send (channel, message)
	Channel channel;
	char * message;
{
	int len;
	if (channel == NULL_CHANNEL)
		return;

	len = MonetStrlen(message);
	if (len > 0)
		send (channel->fid, message, len, 0);
}

extern void Channel_end_of_record (channel)
	Channel channel;
{
	if (channel == NULL_CHANNEL)
		return;
	send (channel->fid, "\n", 1, 0);
}


extern void Channel_flush (channel)
	Channel channel;
{
/* 	flush (channel->fid); */
}

/***************************************************************************/
/*  Get fid of channel  */
/***************************************************************************/

extern int Channel2fid (channel)
	Channel channel;
{
	return channel->fid;
}

extern Channel fid2Channel (fid)
	int fid;
{
	Channel c;

	FORALL(c, ChannelList, next)
		if (c->fid == fid)
			return c;

	sys_error ("No channel with fid %d", fid);
}


/***************************************************************************/
/***************************************************************************/
/*  $Header$  */
/***************************************************************************/
/***************************************************************************/


