#ifndef MONET_QUERY_CLASS_H
#define MONET_QUERY_CLASS_H

#include "db/Entity.h"
#include "query/Lib.h"
#include "general/Channel.h"

typedef caddr_t SubValue;

/***************************************************************************/
/*  There are a number of subclasses of managed object, one for each type  */
/*  of query/protocol.  The definition references a set of virtual         */
/*  functions for handling this subclass.                                  */
/***************************************************************************/

struct query {
	Query next_instance;
	Entity entity;
	caddr_t		subclass_value;	/* Data private to the subclass        */
#define SUBCLASSVALUE(o)	(o->subclass_value)
#define SETSUBCLASSVALUE(o,t)	(o->subclass_value = (String*) MonetNew (t))
	QueryClass	virtual;	/* Virtual functions for subclasses    */
#define VIRTUAL(o)		(o->virtual)

	/***************************************************************************/
	/*  Minimum period between queries  */
	/***************************************************************************/

	TTime		frequency;
	TTime		time_of_last_query;

	/***************************************************************************/
	/*  Monitoring/Querying Statistics                                         */
	/***************************************************************************/

	Cardinal	changes;		/* Status changes per time period */
	Cardinal	change_period;
	Cardinal	pings_since_change;

	/***************************************************************************/
	/*  What to do with the entity                                             */
	/***************************************************************************/

	Boolean		keep;		/* Don't delete after last request released from it */
	Priority	priority;	/*  Priority, when to ping it */
};

/***************************************************************************/
/*  Protocol library definitions,  set of routines for manipulating        */
/*  subclasses of Entity.  Standard functions for use with all protocols.  */
/*  When a new query type is created then a new one of these must be       */
/*  defined.                                                               */
/*  These routines should use the routines in "channel.h" for listening    */
/*  on sockets.                                                            */
/***************************************************************************/

struct query_protocol {
	Query instances;
#define FORINSTANCES(i,class) FORALL (i,class->instances, next_instance);

	Boolean enabled; 	/* Queries enabled */
	char *	title;		/*  Used to identify requests  */
	QueryClass next;

#ifdef MONET_PROTOTYPES
	void	(* finish)		(void);
	SubValue(* create)		(Query, String);
	Query	(* find)		(String);
	void	(* free)		(Query);
	Boolean	(* request)		(Query);
	String	(* as_text)		(Query);

#else /*  MONET_PROTOTYPES */

	void	(* finish)		();
	SubValue(* create)		();
	Query	(* find)		();
	void	(* free)		();
	Boolean	(* request)		();
	String	(* as_text)		();

#endif /*  MONET_PROTOTYPES */

};

/***************************************************************************/
/*  Constructor and destructor for query classes                           */
/***************************************************************************/

#define NEW_QUERYCLASS(queryclass, name) { \
	queryclass = create_query_protocol (); \
	queryclass->enabled  = True; \
	queryclass->finish    = finish; \
	queryclass->instances = NULL; \
	queryclass->create    = create; \
	queryclass->find      = find; \
	queryclass->free      = free; \
	queryclass->request   = request; \
	queryclass->as_text   = as_text; \
	queryclass->title     = name; \
}

#define FREE_QUERYCLASS(queryclass) { \
	MonetFree (queryclass->instances); \
	MonetFree (queryclass); \
}

/***************************************************************************/
/*  Function Prototypes  */
/***************************************************************************/

#ifdef MONET_PROTOTYPES

extern QueryClass create_query_protocol (void);

#else

extern QueryClass create_query_protocol ();

#endif

#endif /* _QUERY_CLASS_H_ */

